@@ -74,47 +74,29 @@ class cGAN():
                          2. target data
 
         # Define forward path.
-        gan = cGAN(generator, discriminator)
+        cgan = cGAN(generator, discriminator)
 
         # Define optimizer.
-        cgan.compile(lr_gen=1e-4, lr_disc=1e-4)
+        cgan.compile()
 
         # Start training process.
-        cgan.fit(x_domain, y_domain, EPOCHS, save_ckpt=5)
+        cgan.fit(train_dataset, ...)
 
         # Export both models.
-        gan.export()
+        cgan.export()
 
-        use gan.restore("relative_path_to_logs") to continue training after
+        use cgan.restore("relative_path_to_logs") to continue training after
         a break.
 
     Attributes:
-        generator (keras.model): Keras NN to act as the generator.
-        discriminator (keras.model): Keras NN to act as the discriminator.
-        generator_optimizer (keras.optimizer): Optimization alg for generator.
-        discriminator_optimizer (keras.optimizer): Optimization alg for disc.
-        init_timestamp (str): Acts as a unique folder identifier.
-        logdir (str): Top level logdir.
-        tensorboard (str): Path to tensorboard summary files.
-        ckpt_dir (str): Path to ckpt files.
-        ckpt_manager (tf.train.CheckpointManager): Deletes old checkpoints.
-        checkpoint (tf.train.Checkpoint): Groups trackable objects, saving and
-            restoring them.
+        generator: keras.model. Keras NN to act as the generator.
+        discriminator: keras.model. Keras NN to act as the discriminator.
+        l1_weight: Integer. Weight coefficient for l1 loss to total
+            loss. Default value from the original paper.
     """
-    def __init__(
-            self,
-            generator,
-            discriminator,
-            l1_weight=100,
-    ):
+    def __init__(self, generator, discriminator, l1_weight=100):
         """Takes two keras.models that take part in the minmax game.
         Both models define the hypothesis set to our learning model.
-
-        Args:
-            generator (keras.model): Keras NN to act as the generator.
-            discriminator (keras.model): Keras NN to act as the discriminator.
-            l1_weight (int, optional): Weight coefficient for l1 loss to total
-                loss. Default value from the original paper.
         """
         self.generator = generator
         self.discriminator = discriminator
@@ -129,20 +111,55 @@ class cGAN():
         self.ckpt_manager = None
         self.checkpoint = None
 
-    def compile(self, optimizer="adam", lr_gen=1e-4, lr_disc=1e-4):
+    def compile(self,
+                optimizer="adam",
+                lr_gen=1e-4,
+                lr_disc=4e-4,
+                loss_scaling=False):
         """Defines the optimization part of the learning algorithm to our
-        learning model.
+        learning model. We propose using separate learning rates (TTUR) for the
+        generator and the discriminator. We propose using TTUR specifically to
+        compensate for the problem of slow learning in a regularized
+        discriminator.
 
         Args:
-            optimizer (str, optional): Optimizer. Defaults to "adam".
-            lr_gen (Float, optional): Learning rate generator. Defaults to 1e4.
-            lr_disc (Float, optional): Learning rate discriminator.
-                Defaults to 1e4.
+            optimizer: String. keras.optimizer. Defaults to "adam".
+            lr_gen: Float. Learning rate used by the generator. Defaults to 1e4.
+            lr_disc: Float. Learning rate used by the discriminator.
+                Defaults to 4e4.
+            loss_scaling: Boolean. Used with mixed precision.
         """
-        # TODO: more optimizer
-        if optimizer == "adam":
+        if optimizer == "adadelta":
+            self.generator_optimizer = tf.keras.optimizers.Adadelta(lr_gen)
+            self.discriminator_optimizer = tf.keras.optimizers.Adadelta(
+                lr_disc)
+        elif optimizer == "adagrad":
+            self.generator_optimizer = tf.keras.optimizers.Adagrad(lr_gen)
+            self.discriminator_optimizer = tf.keras.optimizers.Adagrad(lr_disc)
+        elif optimizer == "adam":
             self.generator_optimizer = tf.keras.optimizers.Adam(lr_gen)
             self.discriminator_optimizer = tf.keras.optimizers.Adam(lr_disc)
+        elif optimizer == "adamax":
+            self.generator_optimizer = tf.keras.optimizers.Adamax(lr_gen)
+            self.discriminator_optimizer = tf.keras.optimizers.Adamax(lr_disc)
+        elif optimizer == "ftrl":
+            self.generator_optimizer = tf.keras.optimizers.Ftrl(lr_gen)
+            self.discriminator_optimizer = tf.keras.optimizers.Ftrl(lr_disc)
+        elif optimizer == "nadam":
+            self.generator_optimizer = tf.keras.optimizers.Nadam(lr_gen)
+            self.discriminator_optimizer = tf.keras.optimizers.Nadam(lr_disc)
+        elif optimizer == "rmsprop":
+            self.generator_optimizer = tf.keras.optimizers.RMSprop(lr_gen)
+            self.discriminator_optimizer = tf.keras.optimizers.RMSprop(lr_disc)
+        elif optimizer == "sgd":
+            self.generator_optimizer = tf.keras.optimizers.SGD(lr_gen)
+            self.discriminator_optimizer = tf.keras.optimizers.SGD(lr_disc)
+
+        if loss_scaling:
+            self.generator_optimizer = tf.keras.mixed_precision.LossScaleOptimizer(
+                self.generator_optimizer)
+            self.discriminator_optimizer = tf.keras.mixed_precision.LossScaleOptimizer(
+                self.discriminator_optimizer)
 
         if self.checkpoint is None:
             self.checkpoint = tf.train.Checkpoint(
@@ -155,38 +172,46 @@ class cGAN():
                                                            max_to_keep=5)
 
     def fit(self,
-            source_domain,
-            target_domain,
-            epochs,
-            batch_size,
-            len_dataset,
+            x=None,
+            y=None,
+            batch_size=None,
+            epochs=1,
+            len_dataset=None,
             save_ckpt=5,
             verbose=1,
             max_outputs=2,
             initial_step=0,
             mlflow=False):
-        """Trains both models for n EPOCHS. Saves ckpts every n EPOCHS.
+        """Trains both models for a fixed number of epochs.
         The training loop together with the optimization algorithm define the
         learning algorithm.
 
         Args:
-            input_domain (tf.dataset): Images that generator sees as input with
-                shape(None, width, height, depth).
-            target_domain (tf.dataset): Images that generator tries to translate
-            to with shape(None, width, height, depth).
-            epochs (int): Number of epochs.
-            save_ckpt (int): Save ckpts every n Epochs.
-            verbose (int, optional): Keras Progbar verbose lvl. Defaults to 1.
-            max_outputs (int, optional): Number of images shown in TB.
+            x: Input data. It could be:
+              - A "tf.data" dataset. Should return a tuple of
+                "(inputs, targets)" or "(inputs)" if y is given.
+            y: Target data. If x is a dataset returning "(inputs)", y should
+                return "(targets)".
+            batch_size: Integer.
+                Number of samples per gradient update, only used to build a
+                keras progressbar. Datasets should generate batches.
+            epochs: Integer. Number of epochs to train the model. An epoch is an
+                iteration over the entire "x" and "y" data provided.
+            save_ckpt: Integer. Save ckpts every n Epochs.
+            verbose: 0, 1, or 2. Verbosity mode.
+                0 = silent, 1 = progress bar, 2 = one line per epoch". Defaults
+                to 1.
+            max_outputs: Integer.  Number of images shown in TB per step.
                 Defaults to 2.
             initial_step (int): Step at which to start training. Useful for
                 resuming a previous run.
+            mlflow: Boolean. Triggers mlflow tracking.
 
         Raises:
             ModelNotCompiledError: Raise if model is not compiled.
         """
         if self.generator_optimizer is None:
-            raise ModelNotCompiledError("use conGAN.compile() first.")
+            raise ModelNotCompiledError("use cGAN.compile() first.")
         if mlflow:
             import mlflow  # pylint: disable=import-outside-toplevel
 
@@ -203,11 +228,15 @@ class cGAN():
         for epoch in range(epochs):
             step_float = 0
             start = time.time()
-            for input_images, target_images in zip(source_domain,
-                                                   target_domain):
+            if y is None:
+                train_dataset = x
+            else:
+                train_dataset = zip(x, y)
+
+            for input_images, target_images in train_dataset:
                 imgs = self.train_step(input_images, target_images, step,
                                        file_writer)
-                # TODO Workaround. summary image & gpu usage
+                # Workaround. summary image & gpu usage
                 with file_writer.as_default():  # pylint: disable=not-context-manager
                     tf.summary.image("source", imgs[0], step=step)
                     tf.summary.image("target", imgs[1], step=step)
@@ -224,6 +253,8 @@ class cGAN():
                 print("\nSaving checkpoint for epoch {} at {}".format(
                     epoch + 1, ckpt_save_path))
 
+            #if mlflow: TODO
+            #    mlflow.log_metric("")
             print(" - Epoch {} finished in {} sec\n".format(
                 epoch + 1, int(time.time() - start)))
 
@@ -231,18 +262,17 @@ class cGAN():
         """Restore model weights from the latest checkpoint.
 
         Args:
-            ckpt_path (str): Relative path to ckpt files.
+            ckpt_path: String. Relative path to ckpt files.
 
         Raises:
             ModelNotCompiledError: Raise if model is not compiled.
         """
-
         restore_path = os.path.dirname(ckpt_path)
         self.logdir = restore_path
         self.tensorboard = os.path.join(self.logdir, "tensorboard")
         self.ckpt_dir = os.path.join(self.logdir, "ckpts")
         if self.ckpt_manager is None:
-            raise ModelNotCompiledError("use conGAN.compile() first.")
+            raise ModelNotCompiledError("use cGAN.compile() first.")
         self.ckpt_manager = tf.train.CheckpointManager(self.checkpoint,
                                                        self.ckpt_dir,
                                                        max_to_keep=5)
@@ -257,7 +287,7 @@ class cGAN():
         """Exports the trained models in hdf5 or SavedModel format.
 
         Args:
-            model_format (str, optional): SavedModel or HDF5. Defaults to hdf5.
+            model_format: String. "SavedModel" or "hdf5". Defaults to hdf5.
         """
         model_dir = os.path.join(self.logdir, "models")
         if not os.path.exists(model_dir):
@@ -277,7 +307,6 @@ class cGAN():
         """Decorated function (@tf.function) that creates a callable tensorflow
         graph from a python function.
         """
-
         with file_writer.as_default():
             # Record gradients for generator and discriminator for each seperate
             # training step.
