@@ -15,10 +15,16 @@
 """The code is written using the Keras Sequential API with a tf.GradientTape
 training loop. This module implements the minmax game used to train a cylce GAN.
 
-Generative Adversarial Networks (GANs) are one of the most interesting ideas in
-computer science today. Two models are trained simultaneously by an adversarial
-process. A generator("the artist") learns to create images that look real, while
-a discriminator ("the art critic") learns to tell real images apart from fakes.
+Image-to-image translation is a class of vision and graphics problems where the
+goal is to learn the mapping between an input image and an output image using a
+training set of aligned image pairs. However, for many tasks, paired training
+data will not be available. We present an approach for learning to translate an
+image from a source domain X to a target domain Y in the absence of paired
+examples. Our goal is to learn a mapping G : X → Y such that the distribution of
+images from G(X) is indistinguishable from the distribution Y using an
+adversarial loss. Because this mapping is highly under-constrained, we couple it
+with an inverse mapping F : Y → X and introduce a cycle consistency loss to
+enforce F(G(X)) ≈ X (and vice versa).
 """
 import os
 import time
@@ -31,26 +37,46 @@ from dnnlab.errors.dnnlab_exceptions import ModelNotCompiledError
 class CycleGAN():
     """Implements a cylce gan learning model.
 
+    Typical usage example:
+
+        generator_target -> keras.model with input layer = dim input data &
+                     output same dim than target data.
+        generator_source -> keras.model with input layer = dim target data &
+                     output same dim than source data.
+
+        discriminator_source -> keras.model with two inputs:
+                         1. input data
+                         2. target data
+        discriminator_target -> keras.model with two inputs:
+                         1. input data
+                         2. target data
+
+        # Define forward path.
+        cycle_gan = CycleGAN(generator_target,
+                             generator_source,
+                             discriminator_target,
+                             discriminator_source)
+
+        # Define optimizer.
+        cycle_gan.compile()
+
+        # Start training process.
+        cycle_gan.fit(train_dataset, ...)
+
+        # Export 4 models.
+        cycle_gan.export()
+
+        use cycle_gan.restore("relative_path_to_logs") to continue training after
+        a break.
+
     Attributes:
-        generator_target (keras.model): Generates target images from source.
-        generator_source(keras.model): Generates source images from targets.
-        discriminator_target (keras.model): Compares fake targets to real ones.
-        discriminator_source(keras.model): Compares fake sources to real ones.
-        generator_target_optimizer (keras.optimizer): Optimization alg for
-            generator_target.
-        generator_source_optimizer (keras.optimizer): Optimization alg for
-            generator_source.
-        discriminator_target_optimizer (keras.optimizer): Optimization alg for
-            discriminator_target.
-        discriminator_source_optimizer (keras.optimizer): Optimization alg for
-            discriminator_source.
-        init_timestamp (str): Acts as a unique folder identifier.
-        logdir (str): Top level logdir.
-        tensorboard (str): Path to tensorboard summary files.
-        ckpt_dir (str): Path to ckpt files.
-        ckpt_manager (tf.train.CheckpointManager): Deletes old checkpoints.
-        checkpoint (tf.train.Checkpoint): Groups trackable objects, saving and
-            restoring them.
+        generator_target: keras.model. Generates target images from source.
+        generator_source: keras.model. Generates source images from targets.
+        discriminator_target: keras.model. Compares fake targets to real ones.
+        discriminator_source: keras.model. Compares fake sources to real ones.
+        alpha_disc: Float. Weight for adversarial loss in total loss.
+        alpha_identity: Float. Weight for identity loss in total loss.
+        lambda_cycle: Float. Weight for cycle loss in total loss.
     """
     def __init__(self,
                  generator_target,
@@ -61,12 +87,6 @@ class CycleGAN():
                  alpha_identity=0.5,
                  lambda_cycle=10):
         """Takes four keras.models that take part in the cycling gan game.
-
-        Args:
-            generator_target (keras.model): Generates target imgs from sources.
-            generator_source(keras.model): Generates source imgs from targets.
-            discriminator_target(keras.model): Compares fake/real targets.
-            discriminator_source(keras.model): Compares fake/real sources.
         """
         self.generator_target = generator_target
         self.generator_source = generator_source
@@ -91,19 +111,42 @@ class CycleGAN():
                 optimizer="adam",
                 lr_gen_target=1e-4,
                 lr_gen_source=1e-4,
-                lr_disc_target=1e-4,
-                lr_disc_source=1e-4):
+                lr_disc_target=4e-4,
+                lr_disc_source=4e-4,
+                loss_scaling=False):
         """Defines the optimization part of the learning algorithm to our
-        learning model.
+        learning model. We propose using separate learning rates (TTUR) for the
+        generator and the discriminator. We propose using TTUR specifically to
+        compensate for the problem of slow learning in a regularized
+        discriminator.
 
         Args:
-            optimizer (str, optional): [description]. Defaults to "adam".
-            lr_gen_g ([type], optional): [description]. Defaults to 1e4.
-            lr_gen_f ([type], optional): [description]. Defaults to 1e4.
-            lr_disc_y ([type], optional): [description]. Defaults to 1e4.
-            lr_disc_x ([type], optional): [description]. Defaults to 1e4.
+            optimizer: String. Defaults to "adam".
+            lr_gen_g: Float. Defaults to 1e4.
+            lr_gen_f: Float. Defaults to 1e4.
+            lr_disc_y: Float. Defaults to 4e4.
+            lr_disc_x: Float. Defaults to 4e4.
+            loss_scaling: Boolean. Used with mixed precision.
         """
-        if optimizer == "adam":
+        if optimizer == "adadelta":
+            self.generator_target_optimizer = tf.keras.optimizers.Adadelta(
+                lr_gen_target)
+            self.generator_source_optimizer = tf.keras.optimizers.Adadelta(
+                lr_gen_source)
+            self.discriminator_target_optimizer = tf.keras.optimizers.Adadelta(
+                lr_disc_target)
+            self.discriminator_source_optimizer = tf.keras.optimizers.Adadelta(
+                lr_disc_source)
+        elif optimizer == "adagrad":
+            self.generator_target_optimizer = tf.keras.optimizers.Adagrad(
+                lr_gen_target)
+            self.generator_source_optimizer = tf.keras.optimizers.Adagrad(
+                lr_gen_source)
+            self.discriminator_target_optimizer = tf.keras.optimizers.Adagrad(
+                lr_disc_target)
+            self.discriminator_source_optimizer = tf.keras.optimizers.Adagrad(
+                lr_disc_source)
+        elif optimizer == "adam":
             self.generator_target_optimizer = tf.keras.optimizers.Adam(
                 lr_gen_target)
             self.generator_source_optimizer = tf.keras.optimizers.Adam(
@@ -112,6 +155,61 @@ class CycleGAN():
                 lr_disc_target)
             self.discriminator_source_optimizer = tf.keras.optimizers.Adam(
                 lr_disc_source)
+        elif optimizer == "adamax":
+            self.generator_target_optimizer = tf.keras.optimizers.Adamax(
+                lr_gen_target)
+            self.generator_source_optimizer = tf.keras.optimizers.Adamax(
+                lr_gen_source)
+            self.discriminator_target_optimizer = tf.keras.optimizers.Adamax(
+                lr_disc_target)
+            self.discriminator_source_optimizer = tf.keras.optimizers.Adamax(
+                lr_disc_source)
+        elif optimizer == "ftrl":
+            self.generator_target_optimizer = tf.keras.optimizers.Ftrl(
+                lr_gen_target)
+            self.generator_source_optimizer = tf.keras.optimizers.Ftrl(
+                lr_gen_source)
+            self.discriminator_target_optimizer = tf.keras.optimizers.Ftrl(
+                lr_disc_target)
+            self.discriminator_source_optimizer = tf.keras.optimizers.Ftrl(
+                lr_disc_source)
+        elif optimizer == "nadam":
+            self.generator_target_optimizer = tf.keras.optimizers.Nadam(
+                lr_gen_target)
+            self.generator_source_optimizer = tf.keras.optimizers.Nadam(
+                lr_gen_source)
+            self.discriminator_target_optimizer = tf.keras.optimizers.Nadam(
+                lr_disc_target)
+            self.discriminator_source_optimizer = tf.keras.optimizers.Nadam(
+                lr_disc_source)
+        elif optimizer == "rmsprop":
+            self.generator_target_optimizer = tf.keras.optimizers.RMSprop(
+                lr_gen_target)
+            self.generator_source_optimizer = tf.keras.optimizers.RMSprop(
+                lr_gen_source)
+            self.discriminator_target_optimizer = tf.keras.optimizers.RMSprop(
+                lr_disc_target)
+            self.discriminator_source_optimizer = tf.keras.optimizers.RMSprop(
+                lr_disc_source)
+        elif optimizer == "sgd":
+            self.generator_target_optimizer = tf.keras.optimizers.SGD(
+                lr_gen_target)
+            self.generator_source_optimizer = tf.keras.optimizers.SGD(
+                lr_gen_source)
+            self.discriminator_target_optimizer = tf.keras.optimizers.SGD(
+                lr_disc_target)
+            self.discriminator_source_optimizer = tf.keras.optimizers.SGD(
+                lr_disc_source)
+
+        if loss_scaling:
+            self.generator_target_optimizer = tf.keras.mixed_precision.LossScaleOptimizer(
+                self.generator_target_optimizer)
+            self.generator_source_optimizer = tf.keras.mixed_precision.LossScaleOptimizer(
+                self.generator_source_optimizer)
+            self.discriminator_target_optimizer = tf.keras.mixed_precision.LossScaleOptimizer(
+                self.discriminator_target_optimizer)
+            self.discriminator_source_optimizer = tf.keras.mixed_precision.LossScaleOptimizer(
+                self.discriminator_source_optimizer)
 
         if self.checkpoint is None:
             self.checkpoint = tf.train.Checkpoint(
@@ -130,11 +228,11 @@ class CycleGAN():
                                                            max_to_keep=5)
 
     def fit(self,
-            source_domain,
-            target_domain,
-            epochs,
-            batch_size,
-            len_dataset,
+            x=None,
+            y=None,
+            batch_size=None,
+            epochs=None,
+            len_dataset=None,
             save_ckpt=5,
             verbose=1,
             max_outputs=2,
@@ -145,17 +243,25 @@ class CycleGAN():
         learning algorithm.
 
         Args:
-            source_domain (tf.dataset): Images that generator sees as source
-                with shape(None, width, height, depth).
-            target_domain (tf.dataset): Images that generator tries to translate
-                to with shape(None, width, height, depth).
-            epochs (int): Number of epochs.
-            save_ckpt (int): Save ckpts every n Epochs.
-            verbose (int, optional): Keras Progbar verbose lvl. Defaults to 1.
-            max_outputs (int, optional): Number of images shown in TB.
+            x: Input data. It could be:
+              - A "tf.data" dataset. Should return a tuple of
+                "(inputs, targets)" or "(inputs)" if y is given.
+            y: Target data. If x is a dataset returning "(inputs)", y should
+                return "(targets)".
+            batch_size: Integer.
+                Number of samples per gradient update, only used to build a
+                keras progressbar. Datasets should generate batches.
+            epochs: Integer. Number of epochs to train the model. An epoch is an
+                iteration over the entire "x" and "y" data provided.
+            save_ckpt: Integer. Save ckpts every n Epochs.
+            verbose: 0, 1, or 2. Verbosity mode.
+                0 = silent, 1 = progress bar, 2 = one line per epoch". Defaults
+                to 1.
+            max_outputs: Integer.  Number of images shown in TB per step.
                 Defaults to 2.
             initial_step (int): Step at which to start training. Useful for
                 resuming a previous run.
+            mlflow: Boolean. Triggers mlflow tracking.
 
 
         Raises:
@@ -179,8 +285,11 @@ class CycleGAN():
         for epoch in range(epochs):
             step_float = 0
             start = time.time()
-            for source_images, target_images in zip(source_domain,
-                                                    target_domain):
+            if y is None:
+                train_dataset = x
+            else:
+                train_dataset = zip(x, y)
+            for source_images, target_images in train_dataset:
                 imgs = self.train_step(source_images, target_images, step,
                                        file_writer)
 
@@ -211,12 +320,11 @@ class CycleGAN():
         """Restore model weights from the latest checkpoint.
 
         Args:
-            ckpt_path (str): Relative path to ckpt files.
+            ckpt_path: String. Relative path to ckpt files.
 
         Raises:
             ModelNotCompiledError: Raise if model is not compiled.
         """
-
         restore_path = os.path.dirname(ckpt_path)
         self.logdir = restore_path
         self.tensorboard = os.path.join(self.logdir, "tensorboard")
@@ -237,7 +345,7 @@ class CycleGAN():
         """Exports the trained models in hdf5 or SavedModel format.
 
         Args:
-            model_format (str, optional): SavedModel or HDF5. Defaults to hdf5.
+            model_format: String. "SavedModel" or "hdf5". Defaults to hdf5.
         """
         model_dir = os.path.join(self.logdir, "models")
         if not os.path.exists(model_dir):
