@@ -91,12 +91,10 @@ discriminators predictions and the correct labels in the binary classification
 task of discriminating real data from fake data.
 
 Both players are trained simultaneously playing the following minimax game:
-    J_D = E_x_from_data [log(D(x))] + E_z_from_generator [log(1-D(G(z)))]
+    J_D = E_x_from_data [log D(x)] + E_z_from_generator [log(1-D(G(z)))]
            "Detect real images"              "Detect fake images"
                 [-∞, 0]                             [-∞, 0]
 
-    J_G = - J_D
-
 - discriminator tries to maximize J_D.
 - generator tries to minimize J_D.
 - equilibrium is a saddle point of the discriminator loss.
@@ -108,15 +106,15 @@ Both players are trained simultaneously playing the following minimax game:
 confidence. In this case log(1-D(G(z))) saturates.
 
 
-Non-Saturating Game (Practise)
+Non-Saturating Game (Practise, NS-GAN)
 ------------------------------
-Each player has its own independently parameterised cost function to eliminate
+Each player has its own independently parameterised loss function to eliminate
 early saturation due to a dominant discriminator. Heuristically motivated; the
 generator can still learn even when discriminator successfully rejects all
 generator samples.
 
 Generator maximizes the log-probability of the discriminator being mistaken:
-    max D(G(z))
+    max log D(G(z))
 
 instead of minimizing the minmax game.
 
@@ -210,11 +208,13 @@ class GAN():
         self.ckpt_dir = os.path.join(self.logdir, "ckpts")
         self.ckpt_manager = None
         self.checkpoint = None
+        self.loss = None
 
     def compile(self,
                 optimizer="adam",
                 lr_gen=1e-4,
                 lr_disc=4e-4,
+                loss="vanilla",
                 loss_scaling=False):
         """Defines the optimization part of the learning algorithm to our
         learning model.
@@ -224,8 +224,14 @@ class GAN():
             lr_gen: Float. Learning rate used by the generator. Defaults to 1e4.
             lr_disc: Float. Learning rate used by the discriminator.
                 Defaults to 4e4.
+            loss: String. Choose between vanilla, hinge and wgan version.
             loss_scaling: Boolean. Used with mixed precision.
         """
+        valid = {"vanilla", "hinge", "wgan"}
+        if loss not in valid:
+            raise ValueError("loss mut be one of %r." % valid)
+        self.loss = loss
+
         if optimizer == "adadelta":
             self.generator_optimizer = tf.keras.optimizers.Adadelta(lr_gen)
             self.discriminator_optimizer = tf.keras.optimizers.Adadelta(
@@ -406,12 +412,16 @@ class GAN():
                 disc_target = self.discriminator(target_images, training=True)
                 disc_generated = self.discriminator(fake_target, training=True)
 
-                gen_loss = self.generator_loss(disc_generated)
-                tf.summary.scalar("generator_loss", gen_loss, step=step)
+                gen_loss = self.generator_loss(self.loss, disc_generated)
+                tf.summary.scalar("generator_{}".format(self.loss),
+                                  gen_loss,
+                                  step=step)
 
-                disc_loss = self.discriminator_loss(disc_target,
+                disc_loss = self.discriminator_loss(self.loss, disc_target,
                                                     disc_generated)
-                tf.summary.scalar("discriminator_loss", disc_loss, step=step)
+                tf.summary.scalar("discriminator_{}".format(self.loss),
+                                  disc_loss,
+                                  step=step)
 
             gradients_of_generator = gen_tape.gradient(
                 gen_loss, self.generator.trainable_variables)
@@ -447,7 +457,7 @@ class GAN():
             return imgs
 
     @staticmethod
-    def discriminator_loss(real_output, fake_output):
+    def discriminator_loss(loss, real_output, fake_output):
         """This method quantifies how well the discriminator is able to
         distinguish real images from fakes. It compares the discriminator's
         predictions on real images to an array of 1s, and the discriminator's
@@ -457,18 +467,29 @@ class GAN():
             real_output: Prob. array of discriminator's performance on real data
             fake_output: Prob. array of discriminator's performance on fake data
         """
-        # This method returns a helper function to compute cross entropy loss.
-        cross_entropy = tf.keras.losses.BinaryCrossentropy(from_logits=True)
-        # Compares real images to tensors of 1 -> if real output 1s
-        real_loss = cross_entropy(tf.ones_like(real_output), real_output)
-        # Compares generated images to tensors of 0 -> if generated output 0s
-        fake_loss = cross_entropy(tf.zeros_like(fake_output), fake_output)
+        total_loss = 0
+        if loss == "vanilla":
+            # This method returns a helper function to compute cross entropy loss.
+            cross_entropy = tf.keras.losses.BinaryCrossentropy(
+                from_logits=True)
+            # Compares real images to tensors of 1 -> if real output 1s
+            real_loss = cross_entropy(tf.ones_like(real_output), real_output)
+            # Compares generated images to tensors of 0 -> if generated output 0s
+            fake_loss = cross_entropy(tf.zeros_like(fake_output), fake_output)
+
+        elif loss == "hinge":
+            real_loss = tf.reduce_mean(tf.nn.relu(1.0 - real_output))
+            fake_loss = tf.reduce_mean(tf.nn.relu(1.0 + fake_output))
+
+        elif loss == "wgan":
+            real_loss = -tf.reduce_mean(real_output)
+            fake_loss = tf.reduce_mean(fake_output)
 
         total_loss = real_loss + fake_loss
         return total_loss
 
     @staticmethod
-    def generator_loss(fake_output):
+    def generator_loss(loss, fake_output):
         """The generator's loss quantifies how well it was able to trick the
            discriminator. Intuitively, if the generator is performing well, the
            discriminator will classify the generated images as real (or 1).
@@ -478,6 +499,16 @@ class GAN():
         Args:
             fake_output: Prob. array of discriminator's performance on fake data
         """
-        # This method returns a helper function to compute cross entropy loss.
-        cross_entropy = tf.keras.losses.BinaryCrossentropy(from_logits=True)
-        return cross_entropy(tf.ones_like(fake_output), fake_output)
+        total_loss = 0
+        if loss == "vanilla":
+            # This method returns a helper function to compute cross entropy loss.
+            cross_entropy = tf.keras.losses.BinaryCrossentropy(
+                from_logits=True)
+            total_loss = cross_entropy(tf.ones_like(fake_output), fake_output)
+
+        elif loss == "hinge":
+            total_loss = -tf.reduce_mean(fake_output)
+
+        elif loss == "wgan":
+            total_loss = -tf.reduce_mean(fake_output)
+        return total_loss
