@@ -12,12 +12,16 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 # ==============================================================================
-"""
-CNN
--------------
+"""The code is written using the Keras Model API with a tf.GradientTape
+training loop. This module implements an image-to-image training loop using
+different losses such as MSE, MAE, SSIM.
 
-Loss Functions:
-    - see article: Loss Functions for Image Restoration with Neural Networks.
+In an image-to-image task one can choose between generative models or
+mse-based Models. Generative models such as cGANs cycle GANs use an adversarial
+loss resulting from a minmax game to learn the mapping. In contrast, mse-based
+models solve a numerical optimization problem using gradient descent. MSE-based
+models can choose from a set of different loss functions described in:
+    - Loss Functions for Image Restoration with Neural Networks. (title)
 """
 import logging
 import os
@@ -33,30 +37,41 @@ logging.getLogger("tensorflow").setLevel(logging.FATAL)
 
 
 class NN():
-    """Implements a CNN learning model.
+    """Implements a mse-based CNN learning model used for image-to-image
+    translation.
 
         Typical usage example:
 
+        model -> keras.model transforming inputs to outputs.
+
+        # Define forward path.
+        nn = NN(model)
+
+        # Define backward path. You can choose from different optimizers and
+        # loss functions.
+        nn.compile()
+
+        # Start training process.
+        nn.fit(train_dataset, valid_dataset, ...)
+
+        # Export  model.
+        nn.export()
+
+        use nn.restore("relative_path_to_logs") to continue training after
+        a break.
+
     Attributes:
-        model (keras.model): Neural net (hypothesis set).
-        optimizer(keras.optimizers): Learning algorithm instance.
-        init_timestamp (str): Acts as a unique folder identifier.
-        logdir (str): Top level logdir.
-        tensorboard (str): Path to tensorboard summary files.
-        ckpt_dir (str): Path to ckpt files.
-        ckpt_manager (tf.train.CheckpointManager): Deletes old checkpoints.
-        checkpoint (tf.train.Checkpoint): Groups trackable objects, saving and
-            restoring them.
+        model: keras.model. Keras NN transforming inputs to outputs.
+        alpha. Float between 0,1. Interpolation weight for combined losses.
     """
-    def __init__(self, model):
-        """ Inits a standard backbone model.
-        Args:
-            model (keras.model): Neural net (hypothesis set).
-        """
+    def __init__(self, model, alpha=0.84):
+        """ Inits nn with ckpt managers and loggers for tensorboard. alpha
+        defaults to 0.84 as suggested by the authors."""
         self.model = model
         self.loss = None
         self.loss_objective = None
         self.optimizer = None
+        self.alpha = alpha
         self.init_timestamp = "CNN-" + datetime.now().strftime("%d%m%Y-%H%M%S")
         self.logdir = os.path.join("logs", self.init_timestamp)
         self.tensorboard = os.path.join(self.logdir, "tensorboard")
@@ -66,14 +81,15 @@ class NN():
 
     @staticmethod
     def _ssim_loss(labels, prediction):
-        """[summary]
-
-        Args:
-            labels ([type]): [description]
-            prediction ([type]): [description]
-
-        Returns:
-            [type]: [description]
+        """Structural similarity index. SSIM evaluates images accounting for the
+        fact that the HVS is sensitice to changes in local structure.
+        Experimental results have shown the superiority of SSIM-based indexes
+        over MSE(l2). As a consequence, SSIM has benn widely employed as a
+        metric to evaluate image processing algorithms. Moreover, given that it
+        can be used as a differnatial cost function, ssim is an alternative
+        choice of loss for image-to-image translation. SSIM has a inverse
+        relationship to MSE which means we use DSSIM = 1 - SSIM as a loss
+        function to minimize. TODO validate calculation.
         """
         ssim = tf.reduce_mean(
             tf.image.ssim(prediction,
@@ -82,16 +98,13 @@ class NN():
         dssim = 1 - ssim
         return dssim
 
-    @staticmethod
-    def _ssim_l1_loss(labels, prediction):
-        """[summary]
-
-        Args:
-            labels ([type]): [description]
-            prediction ([type]): [description]
-
-        Returns:
-            [type]: [description]
+    def _ssim_l1_loss(self, labels, prediction):
+        """By design, both MS-SSIM and SSIM are not particularly sensitive to
+        uniform biases. This can cause changes of brightness or shifts of colors
+        , which typically become more dull. However, MS-SSIM and SSIM preserves
+        the contrast in high-frequency regions better that the other loss
+        functions. To caputre the best of characteristics of both l1 and SSIM,
+        we propose to combine them.
         """
         ssim = tf.reduce_mean(
             tf.image.ssim(prediction,
@@ -100,19 +113,17 @@ class NN():
         dssim = 1 - ssim
 
         l1 = tf.keras.losses.MeanAbsoluteError()
-        dssim_l1 = dssim + l1(prediction, labels)
+        dssim_l1 = self.alpha * dssim + (1 - self.alpha) * l1(
+            prediction, labels)
         return dssim_l1
 
-    @staticmethod
-    def _ssim_l2_loss(labels, prediction):
-        """[summary]
-
-        Args:
-            labels ([type]): [description]
-            prediction ([type]): [description]
-
-        Returns:
-            [type]: [description]
+    def _ssim_l2_loss(self, labels, prediction):
+        """By design, both MS-SSIM and SSIM are not particularly sensitive to
+        uniform biases. This can cause changes of brightness or shifts of colors
+        , which typically become more dull. However, MS-SSIM and SSIM preserves
+        the contrast in high-frequency regions better that the other loss
+        functions. To caputre the best of characteristics of both l2 and SSIM,
+        we propose to combine them.
         """
         ssim = tf.reduce_mean(
             tf.image.ssim(prediction,
@@ -121,20 +132,13 @@ class NN():
         dssim = 1 - ssim
 
         l2 = tf.keras.losses.MeanSquaredError()
-        dssim_l2 = dssim + l2(prediction, labels)
+        dssim_l2 = self.alpha * dssim + (1 - self.alpha) * l2(
+            prediction, labels)
         return dssim_l2
 
     @staticmethod
     def _ssim(labels, prediction):
-        """[summary]
-
-        Args:
-            labels ([type]): [description]
-            prediction ([type]): [description]
-
-        Returns:
-            [type]: [description]
-        """
+        """Used to evaluate image quality in evaluate()"""
         ssim = tf.reduce_mean(
             tf.image.ssim(prediction,
                           labels,
@@ -155,21 +159,21 @@ class NN():
         learning model. Choose between following loss functions:
             - l1 = mean absolute error (mae)
             - l2 = mean squared error (mse)
-            - ssim = structural similarity
+            - ssim = structural similarity loss function
             - ssim_l1 = ssim + l1
             - ssim_l2 = ssim + l2
-            - mssim = multiscale structural similarity
-            - mssim_l1 = mssim + l1
-            - mssim_l2 = mssim + l2
+            - TODO mssim = multiscale structural similarity
+            - TODO mssim_l1 = mssim + l1
+            - TODO mssim_l2 = mssim + l2
 
         Args:
-            loss (keras.losses): Lossfunction for specific task. Defaults to l2.
-            optimizer (str, optional): Optimizer. Defaults to "adam".
-            lr (Float, optional): Learning rate. Defaults to 1e4.
-            loss_scaling (boolean, optional): Loss scaling for mixed presicion.
+            loss: String. Loss function to use.
+            optimizer: String. Optimizer. Defaults to "adam".
+            lr: Float. Learning rate. Defaults to 1e4.
+            loss_scaling: Boolean. Loss scaling for mixed presicion.
                 Defaults to False.
-
         """
+        # Used for tb naming.
         self.loss = loss
 
         if self.loss == "l1":
@@ -212,11 +216,11 @@ class NN():
                                                            max_to_keep=5)
 
     def fit(self,
-            training_data,
-            validation_data,
-            epochs,
-            batch_size,
-            len_dataset,
+            training_data=None,
+            validation_data=None,
+            batch_size=None,
+            epochs=1,
+            len_dataset=None,
             save_ckpt=5,
             verbose=1,
             max_outputs=2,
@@ -227,18 +231,26 @@ class NN():
         learning algorithm.
 
         Args:
-            training_data (tf.dataset): tf.Dataset with
-                shape(None, width, height, depth).
-            validation_data (tf.dataset): tf.Dataset with
-                shape(None, width, height, depth).
-            epochs (int): Number of epochs.
-            batch_size (int): Batch length.
-            save_ckpt (int): Save ckpts every n Epochs.
-            verbose (int, optional): Keras Progbar verbose lvl. Defaults to 1.
-            max_outputs (int, optional): Number of images shown in TB.
+            training_data: Training data. It should be:
+              - A "tf.data" dataset. Should return a tuple of
+                "(inputs, targets)"
+            validation_data: Validation data. It should be:
+              - A "tf.data" dataset. Should return a tuple of
+                "(inputs, targets)"
+            batch_size: Integer.
+                Number of samples per gradient update, only used to build a
+                keras progressbar. Datasets should generate batches.
+            epochs: Integer. Number of epochs to train the model. An epoch is an
+                iteration over the entire "x" and "y" data provided.
+            save_ckpt: Integer. Save ckpts every n Epochs.
+            verbose: 0, 1, or 2. Verbosity mode.
+                0 = silent, 1 = progress bar, 2 = one line per epoch". Defaults
+                to 1.
+            max_outputs: Integer.  Number of images shown in TB per step.
                 Defaults to 2.
-            initial_step (int, optional): Initial step for tb output.
-            mlflow(bool, optional): Tracks validation loss as metric.
+            initial_step (int): Step at which to start training. Useful for
+                resuming a previous run.
+            mlflow: Boolean. Triggers mlflow tracking.
 
         Raises:
             ModelNotCompiledError: Raise if model is not compiled.
