@@ -360,8 +360,14 @@ def _loss_class(
     # Indicator function to only weigh class predictions if an object exists.
     class_mask = tf.squeeze(true_box_conf) * lambda_class
     # Use xentropy between class probabilities.
-    loss_class = tf.nn.sparse_softmax_cross_entropy_with_logits(
-        labels=true_box_class, logits=p_box_class)
+    #loss_class = tf.nn.sparse_softmax_cross_entropy_with_logits(
+    #    labels=true_box_class, logits=p_box_class)
+    # Use binary cross entropy for class prediction. Introduced in YOLOv3
+    # Softmax is not neccessary for good performance and hinders some data
+    # distribution with overlapping laberls (i.e. Woman and Person).
+    # A multilabel approach better models the data.
+    loss_class = tf.nn.sigmoid_cross_entropy_with_logits(labels=true_box_class,
+                                                         logits=p_box_class)
     loss_class = tf.reduce_sum(
         tf.cast(loss_class, dtype=tf.float32) *
         tf.cast(class_mask, dtype=tf.float32))
